#include "openengine.h"

oFont::oFont() throw() : height(0) {}

void oFont::LoadFromFiles(IDirect3DDevice8* const& device, 
						  const _TCHAR* const& textureFileName, 
						  const _TCHAR* const& widthsFileName)
{
	// create a texture
	HRESULT hr = D3DXCreateTextureFromFileEx(
		device, 
		textureFileName, 
		D3DX_DEFAULT, 
		D3DX_DEFAULT, 
		1, 
		0, 
		D3DFMT_A8, 
		D3DPOOL_MANAGED, 
		D3DX_DEFAULT, 
		D3DX_DEFAULT, 
		0, 
		0, 
		0, 
		&texture
	);
	if(FAILED(hr))
		throw _T("oFont::LoadFromFiles(): D3DXCreateTexureFromFileEx failed");
	
	// get the texture dimensions for the font height
	D3DSURFACE_DESC desc;
	hr = texture->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw _T("oFont::LoadFromFiles(): GetLevelDesc failed");
	height = static_cast<uchar>(desc.Height / 16);
	
	// load up the character widths
	widths = static_cast<uchar*>(malloc(256));
	if(!widths)
		throw _T("oFont::LoadFromFiles(): malloc failed");

	oFilePtr file = _tfopen(widthsFileName, _T("rb"));
	if(file == 0)
		throw _T("oFont::LoadFromFiles(): _tfopen failed");

	// read the all and close the file
	size_t count = fread(widths, 1, 256, file);
	if(count != 256)
		throw _T("oFont::LoadFromFiles(): fread failed");
}

oTextBuffer::oTextBuffer() : capacity(0), usage(0) {}

void oTextBuffer::Create(oGraphics& graphics, const ulong& capacity_)
{
	// create the vertex buffer
	graphics.CreateVertexBuffer(buffer, capacity_ * 6);
	// reinitialise
	capacity = capacity_;
	Clear();
}

void oTextBuffer::Clear() throw()
{
	usage = 0;
}

void oTextBuffer::Print(const short& x, const short& y, const _TCHAR* const& text, ulong length, const oFont& font)
{
	if(length == 0)
		length = static_cast<ulong>(_tcslen(text));
	if(usage + length > capacity)
		length = capacity - usage;

	const uchar* widths = font.GetWidths();

	const _TCHAR* input = text;
	oGUIVertex* output = buffer.Lock(usage * 6, length * 6);
	usage += length;

	short position = static_cast<short>(x);
	short topRail = static_cast<short>(y + font.GetHeight());
	short bottomRail = static_cast<short>(y);
	oReal textureScale = _R(1) / _R(16 * font.GetHeight());

	for(; length > 0; --length, ++input, output += 6)
	{
		output[0].y = output[3].y = output[5].y = topRail;
		output[1].y = output[2].y = output[4].y = bottomRail;
		output[0].x = output[1].x = output[3].x = position;
		position = static_cast<short>(position + widths[*input]);
		output[2].x = output[4].x = output[5].x = position;

		long u = ((*input) & 0x0F) * font.GetHeight();
		long v =	(((*input) >> 4) & 0x0F) * font.GetHeight();
		// offset by 0.5 texels to put the texture texel centre at the vertex location
		output[1].v = output[2].v = output[4].v = (_R(v) + _R(0.5)) * textureScale; 
		output[0].v = output[3].v = output[5].v = (_R(v + font.GetHeight()) + _R(0.5)) * textureScale; 
		output[0].u = output[1].u = output[3].u = (_R(u) + _R(0.5)) * textureScale;
		output[2].u = output[4].u = output[5].u = (_R(u + widths[*input]) + _R(0.5)) * textureScale;
	}

    buffer.Unlock();
}

void oTextBuffer::DrawPrimitive(oGraphics& graphics) const
{
	if(usage > 0)
	{
		graphics.SetStreamSource(0, buffer);
		graphics->DrawPrimitive(D3DPT_TRIANGLELIST, 0, usage * 2);
		graphics.ReleaseStreamSource(0);
	}
}
