#pragma once

class oWriteOnlyDynamicBufferTraits
{
public:
	static const ulong usageTraits = D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY;
	static const ulong lockTraits = D3DLOCK_DISCARD;
};

class oWriteOnlyStaticBufferTraits
{
public:
	static const ulong usageTraits = D3DUSAGE_WRITEONLY;
	static const ulong lockTraits = 0;
};

template<class V, class Traits = oWriteOnlyDynamicBufferTraits>
class oVertexBuffer : public oComPtr<IDirect3DVertexBuffer8>::Type, public Traits
{
public:
	V* Lock() { return Lock(0, 0); }

	V* Lock(const ulong& startOffset, const ulong& numVerticesToLock)
	{
		if(!p)
			throw _T("oVertexBuffer::Lock(): Access violation");
		// lock (with discard) the given region of the vertex buffer
		V* vertices = 0;
		HRESULT hr = p->Lock(
			startOffset * sizeof(V), 
			numVerticesToLock * sizeof(V), 
			reinterpret_cast<BYTE**>(&vertices), 
			lockTraits
		);
		if(FAILED(hr))
			throw _T("oVertexBuffer::Lock(): Lock failed");
		// return it
		return vertices;
	}

	void Unlock() throw()
	{
		// unlock the vertex buffer
		if(p) p->Unlock();
	}
};

template<class I, class Traits = oWriteOnlyDynamicBufferTraits>
class oIndexBuffer : public oComPtr<IDirect3DIndexBuffer8>::Type, public Traits
{
public:
	I* Lock() { return Lock(0, 0); }

    I* Lock(const ulong& startOffset, const ulong& numIndicesToLock)
	{
		if(!p)
			throw _T("oIndexBuffer::Lock(): Access violation");
		// lock (with discard) the given region of the vertex buffer
		I* indices = 0;
		HRESULT hr = p->Lock(
			startOffset * sizeof(I), 
			numIndicesToLock * sizeof(I), 
			reinterpret_cast<BYTE**>(&indices), 
			lockTraits
		);
		if(FAILED(hr))
			throw _T("oIndexBuffer::Lock(): Lock failed");
		// return it
		return indices;
	}

	void Unlock() throw()
	{
		if(p) p->Unlock();
	}
};