#include "openengine.h"

void oCreateCubeTriangleListVertices(void* const& buffer, const ulong& stride) throw()
{	
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define FOO(X, Y, Z)	buf = reinterpret_cast<float*>(offset);			\
						*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
						offset += stride;
    FOO(-1, -1, -1) FOO(-1,  1, -1) FOO( 1,  1, -1)
    FOO( 1,  1, -1) FOO( 1, -1, -1) FOO(-1, -1, -1)
    FOO(-1, -1, -1) FOO(-1, -1,  1) FOO(-1,  1,  1)
    FOO(-1,  1,  1) FOO(-1,  1, -1) FOO(-1, -1, -1)
    FOO(-1, -1, -1) FOO( 1, -1, -1) FOO( 1, -1,  1)
    FOO( 1, -1,  1) FOO(-1, -1,  1) FOO(-1, -1, -1)

	FOO( 1,  1,  1) FOO(-1,  1,  1) FOO(-1, -1,  1)
    FOO(-1, -1,  1) FOO( 1, -1,  1) FOO( 1,  1,  1)
    FOO( 1,  1,  1) FOO( 1, -1,  1) FOO( 1, -1, -1)
    FOO( 1, -1, -1) FOO( 1,  1, -1) FOO( 1,  1,  1)
    FOO( 1,  1,  1) FOO( 1,  1, -1) FOO(-1,  1, -1)
    FOO(-1,  1, -1) FOO(-1,  1,  1) FOO( 1,  1,  1)
#undef FOO
}

void oCreateCubeTriangleListNormals(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define BAR(X, Y, Z)	for(ulong n = 0; n < 6; ++n)						\
						{													\
							buf = reinterpret_cast<float*>(offset);			\
							*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
							offset += stride;								\
						}
	BAR(0, 0, -1)
	BAR(-1, 0, 0)
	BAR(0, -1, 0)

	BAR(0, 0, 1)
	BAR(1, 0, 0)
	BAR(0, 1, 0)
#undef BAR
}

void oCreateCubeTriangleListTangents(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define BAR(X, Y, Z)	for(ulong n = 0; n < 6; ++n)						\
						{													\
							buf = reinterpret_cast<float*>(offset);			\
							*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
							offset += stride;								\
						}
	BAR(1, 0, 0)
	BAR(0, 1, 0)
	BAR(0, 0, 1)

	BAR(0, -1, 0)
	BAR(0, 0, -1)
	BAR(-1, 0, 0)
#undef BAR
}

void oCreateCubeTriangleListTexcoords(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define FOO(U, V)	buf = reinterpret_cast<float*>(offset);	\
					*buf++ = _R(U); *buf = _R(V);			\
					offset += stride;
#define BAR(U0, U1, V0, V1)	FOO(U0, V1)	\
							FOO(U0, V0)	\
							FOO(U1, V0)	\
							FOO(U1, V0)	\
							FOO(U1, V1)	\
							FOO(U0, V1)
	for(ulong i = 0; i < 6; ++i)
	{
		BAR(0, 1, 0, 1)
	}
#undef BAR
#undef FOO
}

void oCreateCubeTriangleListNormalMapCoords(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define BAR(X, Y)	for(ulong n = 0; n < 6; ++n)				\
					{											\
						buf = reinterpret_cast<float*>(offset);	\
						*buf++ = _R(X); *buf = _R(Y);			\
						offset += stride;						\
					}
	
	BAR(0.625, 0.385)
	BAR(0.125, 0.385)
	BAR(0.385, 0.385)

	BAR(0.625, 0.125)
	BAR(0.125, 0.125)
	BAR(0.385, 0.125)
#undef BAR
}

static inline oVector3 oEvaluateTorus(const oReal& outerRadius, const oReal& innerRadius, const oReal& u, const oReal& v) throw()
{
	return oVector3(
		oRealCos(u * _R(2) * oRealPi) * (outerRadius - innerRadius * oRealSin(v * _R(2) * oRealPi)), 
		-oRealSin(u * _R(2) * oRealPi) * (outerRadius - innerRadius * oRealSin(v * _R(2) * oRealPi)), 
		-innerRadius * oRealCos(v * _R(2) * oRealPi) 
	);
}

static inline oVector3 oEvaluateTorusNormal(const oReal& u, const oReal& v) throw()
{
	return oVector3(
		-oRealCos(u * _R(2) * oRealPi) * oRealSin(v * _R(2) * oRealPi), 
		oRealSin(u * _R(2) * oRealPi) * oRealSin(v * _R(2) * oRealPi), 
		-oRealCos(v * _R(2) * oRealPi)
	);
}

static inline oVector3 oEvaluateTorusTangent(const oReal& u, const oReal& /*v*/) throw()
{
	return oVector3(
		-oRealSin(u * _R(2) * oRealPi), 
		-oRealCos(u * _R(2) * oRealPi), 
		_R(0)
	);
}

static inline oVector3 oEvaluateTorusBiNormal(const oReal& u, const oReal& v) throw()
{
	return oVector3(
		-oRealCos(u * _R(2) * oRealPi) * oRealCos(v * _R(2) * oRealPi), 
		oRealSin(u * _R(2) * oRealPi) * oRealCos(v * _R(2) * oRealPi), 
		oRealSin(v * _R(2) * oRealPi)
	);
}

static inline oVector3 oEvaluateTorusKnot(const oReal& outerRadius, const oReal& innerRadius, 
										  const oReal& knotRadius, 
										  const oReal& p, const oReal& q, const oReal& lenpq, 
										  const oReal& u, const oReal& v) throw()
{
    return oEvaluateTorus(outerRadius, innerRadius, p * u, q * u) + knotRadius * 
		(oRealCos(v * _R(2) * oRealPi) * oEvaluateTorusNormal(p * u, q * u)
		+ (q / lenpq) * oRealSin(v * _R(2) * oRealPi) * oEvaluateTorusBiNormal(p * u, q * u)
		- (p / lenpq) * oRealSin(v * _R(2) * oRealPi) * oEvaluateTorusTangent(p * u, q * u));
}

static inline oVector3 oEvaluateTorusKnotNormal(const oReal& p, const oReal& q, const oReal& lenpq, const oReal& u, const oReal& v) throw()
{
    return oRealCos(v * _R(2) * oRealPi) * oEvaluateTorusNormal(p * u, q * u)
		+ (q / lenpq) * oRealSin(v * _R(2) * oRealPi) * oEvaluateTorusBiNormal(p * u, q * u)
		- (p / lenpq) * oRealSin(v * _R(2) * oRealPi) * oEvaluateTorusTangent(p * u, q * u);
}

static inline oVector3 oEvaluateTorusKnotTangent(const oReal& p, const oReal& q, const oReal& lenpq, const oReal& u, const oReal& /*v*/) throw()
{
	return (p / lenpq) * oEvaluateTorusTangent(p * u, q * u)
		+ (q / lenpq) * oEvaluateTorusBiNormal(p * u, q * u);
}

void oCreateTorusVertices(void* const& buffer, const ulong& stride, 
						  const oReal& outerRadius, const oReal& innerRadius, 
						  const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorus(outerRadius, innerRadius, _R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void oCreateTorusNormals(void* const& buffer, const ulong& stride, 
						 const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorusNormal(_R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void oCreateTorusTangents(void* const& buffer, const ulong& stride, 
						  const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorusTangent(_R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void oCreateTorusTexcoords(void* const& buffer, const ulong& stride, 
						   const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector2*>(offset) = oVector2(_R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void oCreateTorusIndices(ushort* const& buffer, const ushort& offset, 
						 const ulong outerSegments, const ulong& innerSegments) throw()
{
	// we'll add one degenerate triangle each side of the strip so that
	// the whole torus can be drawn in one DrawIndexedPrimitive call
	ushort* buf = buffer;
#define FOO(N, M)	*buf++ = offset + static_cast<ushort>(N);					\
					for(ulong i = 0; i <= outerSegments; ++i)					\
					{															\
						*buf++ = offset + static_cast<ushort>(N + i);			\
						*buf++ = offset + static_cast<ushort>(M + i);			\
					}															\
					*buf++ = offset + static_cast<ushort>(M + outerSegments);

	for(ulong v = 0; v < innerSegments;)
	{
        ulong v0 = v++;
		FOO(v0 * (outerSegments + 1), v * (outerSegments + 1))
	}

#undef FOO
}

void oCreateTorusKnotVertices(void* const& buffer, const ulong& stride, 
							  const oReal& outerRadius, const oReal& innerRadius, const oReal& knotRadius, 
							  const oReal& p, const oReal& q, 
							  const ulong outerSegments, const ulong& innerSegments) throw()
{
	oReal lenpq = oRealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorusKnot(
				outerRadius, innerRadius, knotRadius, 
				p, q, lenpq, _R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}

void oCreateTorusKnotNormals(void* const& buffer, const ulong& stride, 
							 const oReal& p, const oReal& q, 
							 const ulong outerSegments, const ulong& innerSegments) throw()
{
	oReal lenpq = oRealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorusKnotNormal(
				p, q, lenpq, 
				_R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}

void oCreateTorusKnotTangents(void* const& buffer, const ulong& stride,
							  const oReal& p, const oReal& q,
							  const ulong outerSegments, const ulong& innerSegments) throw()
{
	oReal lenpq = oRealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<oVector3*>(offset) = oEvaluateTorusKnotTangent(
				p, q, lenpq, 
				_R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}
