#include "stdafx.h"

struct SomeVertex { float x, y, z, u, v; };

int APIENTRY _tWinMain(HINSTANCE instance,
					   HINSTANCE /*prevInstance*/,
					   LPTSTR /*cmdLine*/,
					   int /*cmdShow*/)
{
	oGraphics graphics; // scoped here so that MessageBox works for caught exceptions (keeps the window alive)
	try
	{
		// show the graphics loader
		oGraphicsLoader loader;
		loader.SetMinPixelShaderVersion(0xffff); // effectively disallow pixel shaders (we don't use them)
		if(!loader.Show(instance))
			return 0;

		// create the graphics device from the selected loader configuration
        graphics.CreateFromLoader(loader, instance, _T("OpenEngine Test"));

		// create an input handler
		oInput input;
		input.Create(instance, graphics.GetTargetWindow());

		// create a vertex buffer
		oVertexBuffer<SomeVertex, oWriteOnlyDynamicBufferTraits> buffer;
		graphics.CreateVertexBuffer(buffer, 100);

		// create a font
		oFont font;
		font.LoadFromFiles(graphics, _T("media/guifont.tga"), _T("media/guifont.w")); 

		// create a text buffer
		oTextBuffer text;
		text.Create(graphics, 400);
		
		// create a GUI shader
		oVertexShader guiShader;
		oCreateGUIVertexShader(graphics, guiShader);

		// loop until the user quits
		for(;;)
		{
			if(graphics.PurgeMessageQueue())
				return 0;

			// render a very exciting blue screen
			graphics->Clear(0, 0, D3DCLEAR_TARGET, 0xff7fa0e0, 1.0f, 0);
			graphics->BeginScene();

			// throw in some text based on the key states
            text.Clear();
			text.Print(10, 10, _T("Press left, up, right or down..."), 0, font);

			short y = 0;
			const short spacing = 18;
			input.GetImmediateKeyboardData();
			if(input.IsKeyDown(DIK_LEFT))
				text.Print(10, 10 + (++y * spacing), _T("left"), 0, font);
			if(input.IsKeyDown(DIK_UP))
				text.Print(10, 10 + (++y * spacing), _T("up"), 0, font);
			if(input.IsKeyDown(DIK_RIGHT))
				text.Print(10, 10 + (++y * spacing), _T("right"), 0, font);
			if(input.IsKeyDown(DIK_DOWN))
				text.Print(10, 10 + (++y * spacing), _T("down"), 0, font);

			graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);

			graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
			graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
			graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

			graphics->SetVertexShader(guiShader);

			float textColour[] = { 0.0f, 0.0f, 0.0f, 1.0f };
			oMatrix4x4 transform;
			oCreateGUITransform(graphics.GetTargetWidth(), graphics.GetTargetHeight(), transform);
			graphics->SetVertexShaderConstant(GUI_TRANSFORM, &transform, 4);
			graphics->SetVertexShaderConstant(GUI_COLOUR, textColour, 1);

			graphics->SetTexture(0, font.GetTexture());
			graphics->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
			graphics->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

			text.DrawPrimitive(graphics);

			graphics->SetTexture(0, 0);

			graphics->EndScene();
			graphics->Present(0, 0, 0, 0);
            
			// stop chugging away on system resources
			WaitMessage();
		}
	}
	catch(_TCHAR* excuse)
	{
		// show the excuse before ritual suicide
		MessageBox(0, excuse, _T("OpenEngine: Commencing harakiri"), MB_OK | MB_ICONERROR | MB_TASKMODAL);
	}

	return 0;
}
